/*
     ``The contents of this file are subject to the Mozilla Public License
     Version 1.1 (the "License"); you may not use this file except in
     compliance with the License. You may obtain a copy of the License at
     http://www.mozilla.org/MPL/

     Software distributed under the License is distributed on an "AS IS"
     basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
     License for the specific language governing rights and limitations
     under the License.

     The Original Code is SylphDS (www.sylphds.net).

     The Initial Developer of the Original Code is Ficedula (web@sylphds.net).
*/

#include <nds.h>
#include <stdio.h>
#include <string>
#include <sstream>

#include "Sylph2D.h"
#include "SylphUtil.h"

SylphWidget::SylphWidget(SylphBackground& owner) : FBack(owner) {
	FFocussed = false;
}

SylphWidget::~SylphWidget() {
}

void SylphWidget::doTouch(bool down) {
}
 
void SylphWidget::setX(int x) {
	FX = x;
}

void SylphWidget::setY(int y) {
	FY = y;
}

void SylphWidget::setWidth(int w) {
	FWidth = w;
}

void SylphWidget::setHeight(int h) {
	FHeight = h;
}

bool SylphWidget::canFocus() {
	return false;
}

void SylphWidget::setFocussed(bool f) { 
	FFocussed = f; 
}

inline void SkipBGRow(u16* &data, int amount) {
	data += amount * 32 * (256/8);
}

inline void SkipBGTile(u16* &data, int amount) {
	data += amount * 32;
}

//Shaded palette: BGR values are:
//	24/11/0			10/0/0
//	16/0/0			4/0/0
//ie. blue varies 24->4 (21 values), green varies 11->0 (12 values), red doesn't vary
//252 possible colours
void GenFFDefPalette(u16* pal) {
	int i,j;
	for (i=4; i<=24; ++i)
		for (j=0; j<=11; ++j)
			pal[j+(i-4)*12+1] = RGB15(0, j, i) | 0x8000; //force alpha on just in case it's needed
}


void FillFFDefChar(u16* charbase) {
	int i,j;
	for (j=0; j<192; ++j) {
		int lb = 24 - (j/24);
		int rb = 10 - (j/31);
		int lg = 11 - (j/17);
		int rg = 0;
		
		u16 pixels=0;
		for (i=0; i<256; ++i) { 
			u16* block = charbase;
			SkipBGRow(block, j>>3);
			SkipBGTile(block, i>>3);
			
			int bamount = lb - (lb-rb)*i/255;
			int gamount = lg - (lg-rg)*i/255;
			u8 pixel = gamount + (bamount-4) * 12;
			if (i&1) {
				pixels = pixels | (pixel<<8);
				block[ ((i&7)>>1) + 4*(j&7) ] = pixels;
			} else
				pixels = pixel;
		}
	}
}


inline u8 UnpackBitExpand(u32* src, int &shift, u32 expand) {
	u32 data = src[shift>>4];
	data >>= (shift & 0xf)*2;
	shift++;
	data &= 3;
	data <<= 3;
	return (expand >> data) & 0xff;
}

SylphBackground::SylphBackground(u16* charbase, u16* mapbase) : fcharbase(charbase), fmapbase(mapbase), FPalsUsed(0), FPalCount(0) {
	//member init is all we need
}

void SylphBackground::ClearWidgets(bool ManagedOnly) {
	u32 i;
	SylphWidgets w;
	if (ManagedOnly) {
		for (i=0; i < widgets.size(); ++i)
			if (widgets[i].flags & WIDGET_UNMANAGED)
				w.push_back(widgets[i]);
	} 

	for (i=0; i < widgets.size(); ++i)
		if (!(widgets[i].flags & WIDGET_UNMANAGED))
			delete widgets[i].w;
	widgets.clear();
	
	if (ManagedOnly) widgets = w;
}

SylphBackground::~SylphBackground() {
	ClearWidgets();
	delete[] FPalsUsed;
}

void SylphBackground::InitPals(int palcount, SylphBGPalBank palbank) {
	delete[] FPalsUsed;
	int i;
	FPalsUsed = new int[palcount];
	for (i=0; i < palcount; ++i) FPalsUsed[i] = 256;
	FPalCount = palcount;
	FPalBank = palbank;
}

int SylphBackground::ReservePal(int numcolours) {
	int i;
	for (i=0; i < FPalCount; ++i)
		if (FPalsUsed[i] >= numcolours) {
			int index = 256 - FPalsUsed[i];
			FPalsUsed[i] -= numcolours;
			printf("Reserve %d pal colours\n", numcolours);
			return index | (i << 16);
		}
	return 0;
}

void SylphBackground::LoadPalette(int palette, int offset, int colours, u16* data) {
	u16* dest;
	u16 modeE = VRAM_E_CR, modeF = VRAM_F_CR, modeG = VRAM_G_CR, modeH = VRAM_H_CR;
	
	switch (FPalBank) {
		case bgSimple:		dest = &BG_PALETTE[0]; break; 
		case bgBankE:		dest = &VRAM_E[0]; VRAM_E_CR = VRAM_E_LCD | VRAM_ENABLE; break;
		case bgBankF:		dest = &VRAM_F[0]; VRAM_F_CR = VRAM_F_LCD | VRAM_ENABLE; break;
		case bgBankG:		dest = &VRAM_G[0]; VRAM_G_CR = VRAM_G_LCD | VRAM_ENABLE; break;
		case bgSimpleSub:	dest = &BG_PALETTE_SUB[0]; break;
		case bgExtendSub:	dest = &VRAM_H[0]; VRAM_H_CR = VRAM_H_LCD | VRAM_ENABLE; break;
		default:			return;
	}
	printf("LPal: %d cols into pal %d, OS %d\n", colours, palette, offset);
	memcpy(&dest[256*palette + offset], data, colours*2);
	if (VRAM_E_CR != modeE) VRAM_E_CR = modeE;
	if (VRAM_F_CR != modeF) VRAM_F_CR = modeF;
	if (VRAM_G_CR != modeG) VRAM_G_CR = modeG;
	if (VRAM_H_CR != modeH) VRAM_H_CR = modeH;
}

void SylphBackground::SetTiles(int tileX, int tileY, int tileCount, u8 colour, bool forceoverflow) {
	if (((tileX > 31) || (tileY > 23)) && !forceoverflow) return;
	u16* base = fcharbase;
	base += 32 * (32 * tileY + tileX);
	int i;
	u16 value = colour | (colour << 8);
	for (i=0; i < (tileCount * 32); ++i) *base++ = value;
	//TODO: Use DMA or some other more efficient method!
}

void SylphBackground::ClearTiles(int tileX, int tileY, int tileCount) {
	SetTiles(tileX, tileY, tileCount, 0);
}

void SylphBackground::ClearScreen() {
	int i;
	for (i=0; i < 24; ++i) ClearTiles(0, i, 32);
}

void SylphBackground::DrawLine(int x, int y, int length, bool horiz, u8 colour) {
	//This is a hack to reuse the existing code. Could be more efficient.
	u32* packed = new u32[1+length/16];
	int i;
	for (i=0; i < (1+length/16); ++i) packed[i] = 0;
	if (horiz)
		UnpackBlit(&packed[0], length, 1, 0, x, y, colour);
	else
		UnpackBlit(&packed[0], 1, length, 0, x, y, colour); //stride of zero to not increment each row...hacky!
	delete[] packed;
}


SylphWidget* SylphBackground::AddWidget(SylphWidget* w, int flags) { 
	SylphWidgetReg reg;
	reg.w = w;
	reg.flags = flags;
	widgets.push_back(reg); 
	w->FInit();
	return w;
}

void SylphBackground::RemoveWidget(SylphWidget* w) {
	SylphWidgets::iterator it;
	for (it = widgets.begin(); it != widgets.end(); ++it)
		if ((*it).w == w) {
			if (!((*it).flags & WIDGET_UNMANAGED)) delete (*it).w;
			widgets.erase(it);
			break;
		}
}

void SylphBackground::Step(SylphFrame& frame) {
	u32 i;
	for (i=0; i < widgets.size(); ++i)
		widgets[i].w->FDraw();
}

SylphFrameType SylphBackground::DefaultType() {
	return sfDisplayConstant;
}

void SylphBackground::InitUniqueMap(int palette) {
	int i;
	for (i=0; i < (24*32); ++i)
		fmapbase[i] = i | (palette << 12);
}

void SylphBackground::SetTilePals(u16* map, u32 left, u32 top, u32 width, u32 height, u8 pal) {
	u32 i,j;
	left >>= 3;
	top >>= 3;
	width >>= 3;
	height >>= 3;
	for (i=left; i<(left+width); ++i)
		for (j=top; j<(top+height); ++j)
			map[i + j*32] = (map[i + j*32] & 0xfff) | (pal << 12);
}

void SylphBackground::UnpackBlit(u32* src, u32 width, u32 height, u32 stride, u32 bx, u32 by, u32 expand) {
	if ((bx > 255) || (by > 191)) return;

	u32 i,j;
	int shift = 0;
	u16* dest;
	stride >>= 2;
	
	
	if (bx & 1) { //unaligned destination; deal with specially
		dest = fcharbase;
		SkipBGRow(dest, by>>3);
		SkipBGTile(dest, bx>>3);
		dest += (by&7) * 4;
		dest += (bx&7) >> 1;
		for (j=by; j<(by+height); ++j) {
			shift=0;
			*dest = (*dest & 0xff) | UnpackBitExpand(&src[stride*(j-by)], shift, expand) << 8;
			if ((j&7)==7) {
				SkipBGRow(dest, 1); 
				dest -= 28;
			} else dest += 4;
		}
		shift=1;
		bx++;
		width--;
	}
	
	for (j=by; j<(by+height); ++j) { //main copy: writing aligned words
		int rshift = shift;
		dest = fcharbase;
		SkipBGRow(dest, j>>3);
		SkipBGTile(dest, bx>>3);
		dest += (j&7) * 4;
		dest += (bx&7) >> 1;
		for (i=bx; i<(bx+width); i+=2) {
			u16 data = UnpackBitExpand(&src[stride*(j-by)], rshift, expand);
			if ((i >= (bx+width-2)) && (width&1)) //odd pixel at end
				data = data | (*dest & 0xff00);
			else
				data = data | (UnpackBitExpand(&src[stride*(j-by)], rshift, expand) << 8);
			*dest = data;
			if ((i&7)>=6) dest += 29; else dest++;	
		}
	}
		
}

void SylphBackground::SetTile(u16* tile, int x, int y) {
	memcpy(&fcharbase[(x + (y<<5))*32], tile, 64);
}

void SylphBackground::Blit(u8* src, u32 width, u32 height, u32 stride, u32 bx, u32 by) {
	if ((bx+width)>256) return; //todo: allow larger maps? (not using anything that large yet...)
	if ((by+height)>256) return; 
	if (bx>256) return; //first check could overflow!
	if (by>256) return;
	u32 i,j,k;
	int partx = bx & 7;
	int party = by & 7;
	u32 dx = (8 - partx) & 0x7;
	u32 dy = (8 - party) & 0x7;
	if ((partx==0) && (width<8)) dx = width;
	if ((party==0) && (height<8)) dy = height;

	if (dx) {
		u16* dest = fcharbase;
		SkipBGRow(dest, by>>3);
		SkipBGTile(dest, bx>>3);
		dest += 4 * party;
		for (i=0; i<height; ++i) {
			int idx = partx >> 1;
			if (dx & 1) {
				if (partx) {
					dest[idx] = (dest[idx] & 0xff) | (src[i*stride] << 8);
					idx++;
				} else //single pixel is at END not beginning
					dest[dx>>1] = (dest[dx>>1] & 0xff00) | src[i*stride+dx-1];
			}
			int offset = i*stride+(dx&1);
			for (j=0; j<(dx>>1); ++j) {
				dest[idx++] = src[offset] | (src[offset+1] << 8);
				offset += 2;
			}
			dest += 4;
			if (((party + i) & 7) == 7) {
				SkipBGRow(dest, 1);
				//SkipBGTile(dest, -1);
				dest -= 28; ///????
			}
		}
		if (width < 8) return;
		src += dx;
		width -= dx;
		bx += dx;

		if (width & 7) {
			u32 wdone = (width>>3)*8;
			Blit(&src[wdone], width&7, height, stride, bx+wdone, by);
		}
	}
	
	if (dy) {
		u16* dest = fcharbase;
		SkipBGRow(dest, by>>3);
		SkipBGTile(dest, bx>>3);
		for (i=0; i<(width>>3); ++i) {
			dest += 4*party;
			for (j=0; j<dy; ++j) {
				int index = j*stride + i*8;
				for (k=0; k<4; ++k) {
					dest[4*j+k] = src[index] | (src[index+1]<<8);
					index+=2;
				}
			}
			dest += 4*(8-party);
		}
		src += stride*dy;
		height -= dy;
		by += dy;

		if (height & 7) {
			u32 hdone = (height>>3)*8;
			Blit(&src[stride*hdone], width&(~7), height&7, stride, bx, by+hdone);
		}
	}

	if ((height>>3) && (width>>3)) //we have some block data to copy
		for (i=0; i<(height>>3); ++i) {
			u16* dest = fcharbase;
			SkipBGRow(dest, i+(by>>3));
			SkipBGTile(dest, bx>>3);
			
			u16* wsrc;
			int dofree = 0;
			if ((int)src & 1) {
				dofree++;
				wsrc = new u16[stride*8];
				u8* bsrc = &src[i*8*stride];
				u8* bdst = (u8*)wsrc;
				for (j=0; j<(stride*8); ++j)
					*bdst++ = *bsrc++;
			} else
				wsrc = (u16*)&src[i*8*stride];
				
			for (j=0; j<(width>>3); ++j) {
				for (k=0; k<8; ++k) {
					//memcpy(dest, &src[((i*8)+k)*stride + j*8], 8);
					u16 data[4];
					int index = k*(stride>>1)+j*4;
					data[0] = wsrc[index++];
					data[1] = wsrc[index++];
					data[2] = wsrc[index++];
					data[3] = wsrc[index++];
					dest[0] = data[0];
					dest[1] = data[1];
					dest[2] = data[2];
					dest[3] = data[3];
					dest += 4;
				}
			}
			
			if (dofree) delete[] wsrc;
		}
}

SylphFont::SylphFont(SylphFile& file) : FOffset(0), FGlyphCount(0), FGlyphs(NULL), FHistory() {
	int signature;
	file.Read(&signature, 4);
	file.Read(&FGlyphCount, 2);
	file.Read(&FOffset, 2);
	FGlyphs = new u16*[FGlyphCount];
	
	int i;
	for (i=0; i < FGlyphCount; ++i) {
		file.Seek(8 + 2*i, soStart);
		u16 pos;
		file.Read(&pos, 2);
		file.Seek(pos, soStart);
		u16 width, height, stride;
		file.Read(&width, 2);
		file.Read(&height, 2);
		file.Read(&stride, 2);
		FGlyphs[i] = new u16[(width*height*stride) / 2 + 5];
		FGlyphs[i][0] = width;
		FGlyphs[i][1] = height;
		FGlyphs[i][2] = stride;
		file.Read(&FGlyphs[i][3], width*height*stride);
	}
}

SylphFont::~SylphFont() {
	int i;
	for (i=0; i < FGlyphCount; ++i) delete[] FGlyphs[i];
	delete[] FGlyphs;
}

void SylphFont::ClearHistory(SylphWidget* w) {
	FHistory.erase(w);
}

int SylphFont::TextWidth(const char* str) {
	int result = 0;
	while (*str != 0) {
		int glyph = *str - FOffset;
		if (glyph < FGlyphCount) {
			u16* gdata = FGlyphs[glyph];
			result += gdata[0] + 2;
		}
		str++;
	}
	return result;
}

int SylphFont::WrapTextAt(int limit, const char* str) {
	int result = 0, x=0;
	while (*str != 0) {
		int glyph = *str - FOffset;
		if (glyph < FGlyphCount) {
			u16* gdata = FGlyphs[glyph];
			x += gdata[0] + 2;
		}
		if (x > limit) return result;
		result++;
		str++;
	}
	return result;
}

std::string SylphFont::FindHistory(int x, int y) {
	SylphWordHistory::iterator it;
	for (it = FHistory.begin(); it != FHistory.end(); ++it) {
		SylphWordRec& r = it->second;
		if ((r.x <= x) && (x < r.x+r.w) && (r.y <= y) && (y < r.y+r.h)) return r.word;
	}
	return "";
}

std::string SylphFont::HistoryInfo() {
	std::stringstream ss;
	ss << FHistory.size() << " items ";
	if (FHistory.size()) ss << (FHistory.begin())->second.word;
	return ss.str();
}


int SylphFont::DrawLimit(int limit, std::string& str, SylphBackground &bg, int x, int y, int colours, bool transparency, SylphWidget* history) {
	colours <<= 2;
	int palindex = (3+colours) << 24 | ((2+colours) << 16) | ((1+colours) << 8) | 0;
	if (!transparency) palindex |= colours; //actually set bottom colour; although colour group 0 is always transparent
	limit += x;
	SylphWordRec wr;
	wr.y = y;
	wr.h = 16;
	wr.x = 0;
	wr.w = 0;

	int result=0; int pos=0; int wstart=0;
	while (pos < (int)str.size()) {
		int glyph = str[pos] - FOffset;
		
		if ((wr.x == 0) && (str[pos] > 32)) wr.x = x;
		
		if (glyph < FGlyphCount) {
			u16* gdata = FGlyphs[glyph];
			if ((x+gdata[0]+2) > limit)
				return result; 
			bg.UnpackBlit((u32*)&gdata[4], gdata[0], gdata[1], gdata[2], x, y, palindex);
			x += gdata[0] + 2;
		}
		result++;
		pos++;
		if (history)
			if ((str[pos-1] == 32) || (pos == (int)str.size())) 
				if (pos > wstart+1) {
					int len = pos-wstart-1;
					if (pos == (int)str.size()) ++len;
					
					wr.word = Trim(str.substr(wstart, len));
					wr.w = x-wr.x;
					FHistory.insert(std::make_pair(history, wr));
					wr.x = 0;
					wstart = pos;
				}
		
	}
	return result;
}

void SylphFont::Draw(std::string& str, SylphBackground &bg, int x, int y, int colours, bool transparency, SylphWidget* history) {
	DrawLimit(9999, str, bg, x, y, colours, transparency, history);
}

void SylphGradientMap::GenBankedPalettes(u16* pals) {
//4x4 array of 16x16 palettes
//Effectively a 64x64 gradient
	int i,j;
	for (i=0; i<64; ++i) {
		int lb = 24 - (i/8);
		int rb = 10 - (i/10);
		int lg = 11 - ((i+3)/6);
		int rg = 0;
		
		for (j=0; j<64; ++j) {
			int bamount = lb - (lb-rb)*j/63;
			int gamount = lg - (lg-rg)*j/63;
			u16* pal = &pals[256 * (j>>4) + 256 * 4 * (i>>4)];
			pal[(j&0xf) + 0x10 * (i&0xf)] = RGB15(0, gamount, bamount) | 0x8000;
		}
	}
}

void SylphGradientMap::GenBankedChars(SylphBackground &bg, int offset) {
//64x48 pixel gradient (8x6 tiles)
//Each colour would be pointed to by a (4x3) block of pixels ... TODO: Dither them or something similar?
	int i,j,k;
	for (i=0; i<256; ++i) { //for each colour in palette
		for (j=0; j<2; ++j) { //two words (4 pixels) across
			for (k=0; k<3; ++k) { //three pixels down
				u16* dest = bg.fcharbase;
				dest += 32*offset;
				int x,y;
				x = ((i & 0xf) << 2) + (j<<1);
				y = (i >> 4) * 3 + k;
				dest += 32 * (x >> 3); //skip tiles across
				dest += ((x & 0x7) >> 1); //skip to correct X-pos in tile
				dest += 32 * 8 * (y >> 3); //skip rows down
				dest += 4 * (y & 7); //skip to correct Y-pos in tile
				if (i)
					*dest = i | (i<<8);
				else
					*dest = 1 | (1<<8); //fudge it - palette index 0 == transparent
			}
		}
	}
}

void SylphGradientMap::GenBankedMap(SylphBackground &bg, int offset) {
	u16* map = bg.fmapbase;
	int i,j;
	for (i=0; i<32; ++i)
		for (j=0; j<24; ++j)
			map[i+j*32] = ((i&7) + (j%6)*8 + offset) | (((i>>3) + (j/6)*4) << 12);
}

SylphTiles::SylphTiles(SylphFile& file) : FWidth(0), FHeight(0), FColours(0), FData(0), FPalette(0) {
	printf("SylphTiles::SylphTiles");
	file.Seek(0, soStart);
	printf("!");
	file.Read(&FWidth, 2);
	printf("!");
	file.Read(&FHeight, 2);
	printf("!");
	u16 format, flags;
	printf("!");
	file.Read(&format, 2);
	printf("!");
	file.Read(&flags, 2);
	printf("!");
	int bits;
	printf("!");
	switch(format) {
		case GL_RGB16:	{
						FColours = 16;
						FPalette = new u16[16];
						file.Read(&FPalette[0], FColours*2);
						bits=4;
						break;
						}
		case GL_RGB256:{
						FColours = 256;
						FPalette = new u16[256];
						file.Read(&FPalette[0], FColours*2);
						bits=8;
						break;
						}
		default:		return;
		
	}
	FDataFormat = format;
	FDataSize = (FWidth*FHeight) / (8/bits);
	printf("SylphTiles w/h/c/ds %d/%d/%d/%d", FWidth, FHeight, FColours, FDataSize);
	FData = new u8[FDataSize];
	file.Read(&FData[0], FDataSize);
}

SylphTiles::~SylphTiles() {
	delete[] FData;
	delete[] FPalette;
}

void SylphTiles::LoadPalette(SylphBackground& back, int PalReserve) {
	back.LoadPalette(PalReserve >> 16, PalReserve & 0xff, FColours, &FPalette[0]);
}

void SylphTiles::LoadTiles(u16* dest, int paloffset) {
	switch(FDataFormat) {
		case GL_RGB16:	{
							u8* src = FData;
							u16* dst = dest;
							int i;
							for (i=0; i < FDataSize; ++i) {
								u16 data=0;
								if (*src & 0xf) data = ((*src & 0xf)+paloffset);
								if (*src & 0xf0) data |= ((*src >> 4)+paloffset) << 8;
								*dst++ = data;
								src++;
							}
							break;
						}
		case GL_RGB256:{
							int i;
							u16* dst = dest;
							for (i=0; i < (FDataSize/2); ++i)
								*dst++ = (FData[i*2]+paloffset) | ((FData[i*2+1]+paloffset) << 8);
							break;
						}
		default:		return;
	}
}
