#ifndef SYLPH2D_INCLUDE
#define SYLPH2D_INCLUDE

#include "SylphActions.h"
#include "SylphData.h"
#include <vector>
#include <map>
#include <string>
#include <utility>

class SylphGradientMap;
class SylphBackground;

class SylphWidget {
	protected:
		SylphBackground& FBack;
		int FX;
		int FY;
		int FWidth;
		int FHeight;
		bool FFocussed;
	protected:
		virtual void FInit() = 0;
		virtual void FDraw() = 0;
	public:
		friend class SylphBackground;
		SylphWidget(SylphBackground& owner);
		virtual ~SylphWidget();
		
		inline int getX() { return FX; }
		inline int getY() { return FY; }
		inline int getWidth() { return FWidth; }
		inline int getHeight() { return FHeight; }
		inline bool getFocussed() { return FFocussed; }
		virtual void setFocussed(bool f);
		virtual void setX(int x);
		virtual void setY(int y);
		virtual void setWidth(int w);
		virtual void setHeight(int h);
		virtual void doTouch(bool down);
		virtual bool canFocus();
};

struct SylphWidgetReg {
	SylphWidget* w;
	int flags;
};

typedef std::vector<SylphWidgetReg> SylphWidgets;

static const int WIDGET_UNMANAGED = 0x1;

enum SylphBGPalBank {bgSimple, bgBankE, bgBankF, bgBankG, bgSimpleSub, bgExtendSub};

class SylphBackground: public SylphAction {
	protected:
		u16* fcharbase;
		u16* fmapbase;
		SylphWidgets widgets;
		int* FPalsUsed;
		int FPalCount;
		SylphBGPalBank FPalBank;
	public:
		friend class SylphGradientMap;

		void Step(SylphFrame& frame);
		SylphFrameType DefaultType();
	
		SylphBackground(u16* charbase, u16* mapbase);
		~SylphBackground();

		SylphWidget* AddWidget(SylphWidget* w, int flags=0);
		void RemoveWidget(SylphWidget* w);
		void ClearWidgets(bool ManagedOnly=false);
		
		void SetTilePals(u16* map, u32 left, u32 top, u32 width, u32 height, u8 pal);
		void UnpackBlit(u32* src, u32 width, u32 height, u32 stride, u32 bx, u32 by, u32 expand);
		void Blit(u8* src, u32 width, u32 height, u32 stride, u32 bx, u32 by);
		void SetTile(u16 tile[32], int x, int y);
		inline void SetMap(u16 character, int x, int y) { fmapbase[x+(y<<5)] = character; };
		void SetTiles(int tileX, int tileY, int tileCount, u8 colour, bool forceoverflow = false);
		void ClearTiles(int tileX, int tileY, int tileCount);
		void DrawLine(int x, int y, int length, bool horiz, u8 colour);
		void ClearScreen();
		void InitUniqueMap(int palette);
		void InitPals(int palcount, SylphBGPalBank palbank);
		int ReservePal(int numcolours);
		void LoadPalette(int palette, int offset, int colours, u16* data);
		inline void SetMap(int x, int y, int tile, int pal) { fmapbase[x+y*32] = tile | (pal<<12); }
		inline u16* TileAddress(int x, int y) { return &fcharbase[(x+y*32)*32]; }
};

class SylphTiles {
	protected:
		int FWidth;
		int FHeight;
		int FColours;
		u8* FData;
		u16* FPalette;
		int FDataSize;
		int FDataFormat;
	public:
		SylphTiles(SylphFile& file);
		~SylphTiles();
		
		void LoadPalette(SylphBackground& back, int PalReserve);
		void LoadTiles(u16* dest, int paloffset);
		inline int numColours() { return FColours; }
};

struct SylphWordRec {
	int x;
	int y;
	int w;
	int h;
	std::string word;
};

typedef std::multimap<SylphWidget*, SylphWordRec> SylphWordHistory;

class SylphFont {
	protected:
		int FOffset;
		int FGlyphCount;
		u16** FGlyphs;
		SylphWordHistory FHistory;
	public:
		SylphFont(SylphFile& file);
		~SylphFont();
		void Draw(std::string& str, SylphBackground &bg, int x, int y, int colours=0, bool transparency=true, SylphWidget* history=NULL);
		int DrawLimit(int limit, std::string& str, SylphBackground &bg, int x, int y, int colours=0, bool transparency=true, SylphWidget* history=NULL);
		int TextWidth(const char* str);
		int WrapTextAt(int limit, const char* str);
		void ClearHistory(SylphWidget* w);
		std::string FindHistory(int x, int y);
		std::string HistoryInfo();
};

class SylphGradientMap {
	public:
		void GenBankedPalettes(u16* pals);
		void GenBankedChars(SylphBackground &bg, int offset);
		void GenBankedMap(SylphBackground &bg, int offset);
};

#endif
