/*
     ``The contents of this file are subject to the Mozilla Public License
     Version 1.1 (the "License"); you may not use this file except in
     compliance with the License. You may obtain a copy of the License at
     http://www.mozilla.org/MPL/

     Software distributed under the License is distributed on an "AS IS"
     basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
     License for the specific language governing rights and limitations
     under the License.

     The Original Code is SylphDS (www.sylphds.net).

     The Initial Developer of the Original Code is Ficedula (web@sylphds.net).
*/

#include "SylphLayout.h"
#include "SylphData.h"
#include "SylphUtil.h"
#include "SylphWidgets.h"

#include "nds.h"

#include <stdlib.h>
#include <string>
#include <map>

void SylphLayout::FClear() {
	if (FOwns) {
		SylphNamedWidgets::iterator it;
		for (it = FWidgets.begin(); it != FWidgets.end(); ++it)
			delete it->second;
	}
	FWidgets.clear();
	FLastWidget = NULL;
	FFocus = NULL;
	FLastTouch = ttNone;
}

void SylphLayout::FLoad() {
	FClear();
	unsigned i;
	for (i=0; i < FData.size(); ++i) {
		if (FData[i] == "") continue;
		if (FData[i][0] == '#') continue;
		StringList tokens = Tokenise(FData[i], "\t");
		if (tokens.size() == 0) continue;
		
		int flags = FOwns ? WIDGET_UNMANAGED : 0;
		//TEXT / labels
		if (tokens[0] == "TEXT" && tokens.size() >= 7) {
			int x = atoi(tokens[2].c_str());
			int y = atoi(tokens[3].c_str());
			int w = atoi(tokens[4].c_str());
			int h = atoi(tokens[5].c_str());
			int colour = tokens.size() > 8 ? atoi(tokens[8].c_str()) : 0;
			SylphTextType type = sttNormal;
			if (tokens.size() > 7 && tokens[7] == "DYNAMIC") type = sttDynamic;
			else if (tokens.size() > 7 && tokens[7] == "SCROLLING") type = sttScrolling;
			SylphText* txt = new SylphText(FBack, tokens[6], x, y, w, h, FFont, type, colour);
			FWidgets.insert(make_pair(tokens[1], txt));
			FBack.AddWidget(txt, flags);
		}
		//MEMO / memo text boxes
		else if (tokens[0] == "MEMO" && tokens.size() >= 5) {
			int x = atoi(tokens[2].c_str());
			int y = atoi(tokens[3].c_str());
			int w = atoi(tokens[4].c_str());
			int h = atoi(tokens[5].c_str());
			SylphMemo* memo = new SylphMemo(FBack, FFont, x, y, w, h);
			if (tokens.size() > 6) memo->setColour(atoi(tokens[6].c_str()));
			FWidgets.insert(make_pair(tokens[1], memo));
			FBack.AddWidget(memo, flags);
		}
		//LIST / listboxes
		else if (tokens[0] == "LIST" && tokens.size() >= 6) {
			int x = atoi(tokens[2].c_str());
			int y = atoi(tokens[3].c_str());
			int w = atoi(tokens[4].c_str());
			int h = atoi(tokens[5].c_str());
			SylphList* list = new SylphList(FBack, FFont, x, y, w, h);
			FWidgets.insert(make_pair(tokens[1], list));
			FBack.AddWidget(list, flags);
			int highm = 0, highc = 0, highi = 0, highs = 0;
			if (tokens.size() > 6) highm = atoi(tokens[6].c_str());
			if (tokens.size() > 7) highc = atoi(tokens[7].c_str());
			if (tokens.size() > 8) highi = atoi(tokens[8].c_str());
			if (tokens.size() > 9) highs = atoi(tokens[9].c_str());
			list->setHighlights(highm, highc, highi, highs);
		}
		else if (tokens[0] == "BORDER" && tokens.size() >= 7) {
			int x = atoi(tokens[2].c_str());
			int y = atoi(tokens[3].c_str());
			int w = atoi(tokens[4].c_str());
			int h = atoi(tokens[5].c_str());
			printf("Border at %d/%d/%d/%d %s\n", x, y, w, h, tokens[6].c_str());
			SylphFile* file = Filer().GetData(tokens[6]);
			if (file) {
				SylphBorder* border = new SylphBorder(FBack, x, y, w, h, *file);
				delete file;
				FWidgets.insert(make_pair(tokens[1], border));
				FBack.AddWidget(border, flags);
			} else printf("Border not loaded");
		} else if (tokens[0] == "BUTTON" && tokens.size() >= 7) {
			int x = atoi(tokens[2].c_str());
			int y = atoi(tokens[3].c_str());
			int w = atoi(tokens[4].c_str());
			int h = atoi(tokens[5].c_str());
			SylphButton* btn = new SylphButton(FBack, FFont, x, y, w, h, tokens[6]);
			FWidgets.insert(make_pair(tokens[1], btn));
			FBack.AddWidget(btn, flags);
		} else
			printf("SylphLayout: skipping unknown tokens %s\n", tokens[0].c_str());
	}
}

SylphLayout::SylphLayout(SylphBackground& back, SylphFont& font, bool OwnWidgets) : FBack(back), FFont(font), FOwns(OwnWidgets) {
	FLastTouch = ttNone;
	FLastWidget = NULL;
	FFocus = NULL;
}

SylphLayout::~SylphLayout() {
	FClear();
}

void SylphLayout::Load(SylphFile& file) {
	FData = LoadSL(file);
	FLoad();
}

void SylphLayout::Load(StringList sl) {
	FData = sl;
	FLoad();
}

std::string SylphLayout::getName(SylphWidget* w) {
	SylphNamedWidgets::iterator it;
	for (it = FWidgets.begin(); it != FWidgets.end(); ++it)
		if ((*it).second == w) return (*it).first;
	return "";
}

SylphWidget* SylphLayout::Widget(std::string name) {
	SylphNamedWidgets::iterator it = FWidgets.find(name);
	if (it != FWidgets.end()) 
		return it->second; 
	else 
		return NULL;
}

TouchResult SylphLayout::ProcessTouch() {
	TouchResult r;
	touchPosition txy;
	txy = touchReadXY();
	int tx = (txy.x-280) * 10 / 137;
	int ty = (txy.y-200) * 10 / 187;
	
//	int down = keysDown() & KEY_TOUCH ? 1 : 0;
	
	if (keysHeld() & KEY_TOUCH) {
		if ((FLastTouch == ttDown) || (FLastTouch == ttHold)) {
			r.type = ttHold;
			r.widget = FLastWidget;
			r.name = getName(FLastWidget); //inefficient
		} else {
			r.type = ttDown;
			r.widget = NULL;
			SylphNamedWidgets::iterator it;
			for (it = FWidgets.begin(); it != FWidgets.end(); ++it) {
				SylphWidget* w = it->second;
				if ((w->getX() <= tx) && ((w->getX()+w->getWidth()) >= tx) && (w->getY() <= ty) && ((w->getY()+w->getHeight()) >= ty)) {
					r.widget = w;
					break;
				}
			}
			r.name = getName(r.widget);
			if (r.widget) r.widget->doTouch(true);
		}
	} else {
		if ((FLastTouch == ttDown) || (FLastTouch == ttHold)) {
			r.type = ttUp;
			r.widget = FLastWidget;
			r.name = getName(FLastWidget); //inefficient
			if (r.widget) r.widget->doTouch(false);
		} else {
			r.type = ttNone;
			r.widget = NULL;
			r.name = "";
		}
	}

	FLastTouch = r.type;
	FLastWidget = r.widget;
	return r;
}

inline int D2DSQ(int x1, int y1, int x2, int y2) {
	return (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
}

SylphWidget* SylphLayout::findNextWidget(SylphWidget& from, int vert, int horiz, bool testFocus) {
	if (vert==0 && horiz==0) return NULL;
	int xf, yf, xw, yh;
	xf = from.getX(); xw = from.getWidth();
	yf = from.getY(); yh = from.getHeight();
	int ldist = 999999; 
	SylphWidget* best = NULL;
	SylphNamedWidgets::iterator it;
	for (it = FWidgets.begin(); it != FWidgets.end(); ++it) {
		if (it->second == &from) continue;
		if (testFocus && !it->second->canFocus()) continue;
		if (horiz < 0 && it->second->getX() >= xf) continue;
		if (horiz > 0 && (it->second->getX()+it->second->getWidth()) <= (xf+xw)) continue;
		if (vert < 0 && it->second->getY() >= yf) continue;
		if (vert > 0 && (it->second->getY()+it->second->getHeight()) <= (yf+yh)) continue;
		int dist = D2DSQ(xf, yf, it->second->getX(), it->second->getY());
		int dist2;
		if (horiz) { 
			dist2 = D2DSQ(xf, yf, it->second->getX()+it->second->getWidth(), it->second->getY()); 
			if (dist2 < dist) dist = dist2;
		}
		if (vert) { 
			dist2 = D2DSQ(xf, yf, it->second->getX(), it->second->getY()+it->second->getHeight()); 
			if (dist2 < dist) dist = dist2;
		}
		bool ideal=false;
		if (horiz && ((it->second->getY() + it->second->getHeight()/2) == (yf+yh/2))) ideal=true;
		if (vert && ((it->second->getX() + it->second->getWidth()/2) == (xf+xw/2))) ideal=true;
		if (!ideal) dist *= 2;

		if (dist < ldist) {
			best = it->second;
			ldist = dist;
		}
	}
	return best;
}

/*
	Recognised layout lines:
	#fefefefefef			
		-Comment line
	TEXT	NAME	x	y	w	h	Abcdef...	{DYNAMIC}	{colour}
		-Text/label component. Item name, x/y position, width/height followed by text, optional "DYNAMIC" or "SCROLLING" token, optional colour is font colour group
	LIST	NAME	x	y	width	height	{highlightmethod}	{highlightcolour}	{highlightindent}	{highlightshade}
		-Listbox component. Item name, x/y position, w/h size, optional highlight parameters afterwards.
	BORDER	NAME	x	y	width	height	filename
		-Border component; x/y position, width/height size, filename to load from
	MEMO	NAME	x	y	width	height
		-Memo/texbox component: position and size
	BUTTON	NAME	x	y	width	height	text
		-Clickable button: position, size, text

*/
